/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.common.binaryfile;

import java.io.OutputStream;
import java.util.List;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.stream.Stream;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import com.google.protobuf.ByteString;

import de.ozgcloud.alfa.common.GrpcUtil;
import de.ozgcloud.alfa.common.callcontext.ContextService;
import de.ozgcloud.alfa.common.file.OzgFile;
import de.ozgcloud.alfa.common.file.OzgFileMapper;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.vorgang.grpc.binaryFile.BinaryFileServiceGrpc.BinaryFileServiceBlockingStub;
import de.ozgcloud.vorgang.grpc.binaryFile.BinaryFileServiceGrpc.BinaryFileServiceStub;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcBinaryFilesRequest;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcFindFilesResponse;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcGetBinaryFileDataRequest;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileMetaData;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileRequest;
import de.ozgcloud.vorgang.grpc.file.GrpcOzgFile;
import net.devh.boot.grpc.client.inject.GrpcClient;

@Service
class BinaryFileRemoteService {

	@GrpcClient(GrpcUtil.VORGANG_MANAGER_GRPC_CLIENT)
	private BinaryFileServiceBlockingStub serviceStub;
	@GrpcClient(GrpcUtil.VORGANG_MANAGER_GRPC_CLIENT)
	private BinaryFileServiceStub asyncServiceStub;
	@Autowired
	private OzgFileMapper mapper;
	@Autowired
	private FileIdMapper fileIdMapper;

	@Autowired
	private ContextService contextService;

	static final int CHUNK_SIZE = 255 * 1024;

	public OzgFile getFile(FileId fileId) {
		var response = serviceStub.findBinaryFilesMetaData(buildGrpcBinaryFileRequest(fileId));

		return mapOzgFile(response.getFile(0));
	}

	private OzgFile mapOzgFile(GrpcOzgFile file) {
		return mapper.toFile(file);
	}

	GrpcBinaryFilesRequest buildGrpcBinaryFileRequest(FileId fileId) {
		return GrpcBinaryFilesRequest.newBuilder()
				.addFileId(fileId.toString())
				.setContext(contextService.createCallContext())
				.build();
	}

	public CompletableFuture<FileId> uploadFile(UploadBinaryFileRequest uploadBinaryFileRequest) {
		var fileIdFuture = new CompletableFuture<FileId>();
		var responseObserver = createUploadBinaryFileObserver(fileIdFuture, uploadBinaryFileRequest);

		asyncServiceStub.uploadBinaryFileAsStream(responseObserver);

		return fileIdFuture;
	}

	BinaryFileUploadStreamObserver createUploadBinaryFileObserver(CompletableFuture<FileId> fileIdFuture,
			UploadBinaryFileRequest uploadBinaryFileRequest) {
		var metadataRequest = buildMetaDataRequest(uploadBinaryFileRequest);
		var streamer = new ChunkedFileSender<>(uploadBinaryFileRequest.getUploadStream(), CHUNK_SIZE, this::buildChunkRequest, metadataRequest);
		return new BinaryFileUploadStreamObserver(fileIdFuture, streamer);
	}

	GrpcUploadBinaryFileRequest buildMetaDataRequest(UploadBinaryFileRequest uploadBinaryFileRequest) {
		return GrpcUploadBinaryFileRequest.newBuilder()
				.setMetadata(GrpcUploadBinaryFileMetaData.newBuilder()
						.setContext(contextService.createCallContext())
						.setVorgangId(uploadBinaryFileRequest.getVorgangId())
						.setField(uploadBinaryFileRequest.getField())
						.setContentType(uploadBinaryFileRequest.getContentType())
						.setSize(uploadBinaryFileRequest.getSize())
						.setFileName(uploadBinaryFileRequest.getName()))
				.build();
	}

	GrpcUploadBinaryFileRequest buildChunkRequest(byte[] bytes) {
		return GrpcUploadBinaryFileRequest.newBuilder().setFileContent((ByteString.copyFrom(bytes))).build();
	}

	public Stream<OzgFile> getFiles(List<FileId> fileIds) {
		var response = serviceStub.findBinaryFilesMetaData(buildBinaryFilesRequest(fileIds));

		return mapGrpcFindFilesRespones(response);
	}

	private Stream<OzgFile> mapGrpcFindFilesRespones(GrpcFindFilesResponse response) {
		return response.getFileList().stream().map(mapper::toFile);
	}

	GrpcBinaryFilesRequest buildBinaryFilesRequest(List<FileId> fileIds) {
		return GrpcBinaryFilesRequest.newBuilder()
				.addAllFileId(mapFileIds(fileIds))
				.setContext(contextService.createCallContext())
				.build();
	}

	private List<String> mapFileIds(List<FileId> fileIds) {
		return fileIds.stream().map(fileIdMapper::toString).toList();
	}

	public void writeFileContent(FileId fileId, OutputStream out) {
		var streamFuture = new CompletableFuture<Boolean>();
		var responseObserver = createDownloadBinaryFileObserver(streamFuture, out);

		asyncServiceStub.getBinaryFileContent(buildGrpcGetBinaryFileDataRequest(fileId), responseObserver);

		waitUntilFutureToComplete(streamFuture);
	}

	void waitUntilFutureToComplete(CompletableFuture<Boolean> streamFuture) {
		try {
			streamFuture.get(10, TimeUnit.MINUTES);
		} catch (InterruptedException e) {
			Thread.currentThread().interrupt();
			throw new TechnicalException(e.getMessage(), e);
		} catch (ExecutionException | TimeoutException e) {
			throw new TechnicalException(e.getMessage(), e);
		}
	}

	BinaryFileDownloadStreamObserver createDownloadBinaryFileObserver(CompletableFuture<Boolean> streamFuture, OutputStream out) {
		return new BinaryFileDownloadStreamObserver(streamFuture, out);
	}

	private GrpcGetBinaryFileDataRequest buildGrpcGetBinaryFileDataRequest(FileId fileId) {
		return GrpcGetBinaryFileDataRequest.newBuilder()
				.setContext(contextService.createCallContext())
				.setFileId(fileIdMapper.toString(fileId)).build();
	}
}