/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.common.downloadtoken;

import static de.ozgcloud.alfa.JwtTokenUtil.*;
import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collection;
import java.util.List;
import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;

import com.auth0.jwt.exceptions.JWTVerificationException;

import de.ozgcloud.alfa.JwtTokenUtil;
import de.ozgcloud.alfa.common.binaryfile.FileId;
import de.ozgcloud.alfa.common.user.CurrentUserService;
import de.ozgcloud.alfa.common.user.UserProfile;
import de.ozgcloud.alfa.common.user.UserProfileTestFactory;
import de.ozgcloud.common.errorhandling.TechnicalException;
import io.jsonwebtoken.Claims;
import jakarta.servlet.http.HttpServletRequest;
import jakarta.servlet.http.HttpServletResponse;

class DownloadTokenServiceTest {

	@InjectMocks
	private DownloadTokenService service;
	@Mock
	private CurrentUserService userService;
	@Mock
	private JwtTokenUtil jwtTokenUtil;
	@Mock
	private HttpServletRequest request;
	@Mock
	private HttpServletResponse response;

	private final static String FAKE_TOKEN = "xxx";
	private final static Collection<String> ORGE_IDS = List.of("258994");

	@Nested
	class TestCreateToken {

		final FileId fileId = FileId.createNew();

		final UserProfile user = UserProfileTestFactory.create();

		@BeforeEach
		void mockUserService() {
			when(userService.getUser()).thenReturn(user);
		}

		@Test
		void shouldCallUserService() {
			service.createToken(fileId);

			verify(userService).getUser();
		}

		@Test
		void shouldCallJwtTokenUtil() {
			service.createToken(fileId);

			verify(jwtTokenUtil).generateToken(fileId, user);
		}
	}

	@Nested
	class TestHandleToken {

		@Test
		void shouldVerifyTokenIfPresent() throws Exception {
			mockClaims(FIRSTNAME_CLAIM, LASTNAME_CLAIM);

			handleToken();

			verify(jwtTokenUtil).verifyToken(FAKE_TOKEN);
		}

		@Test
		void shouldGetUserFromTokenWithoutUserFirstname() {
			mockClaims(null, FIRSTNAME_CLAIM);

			UserProfile user = service.getUserFromToken(FAKE_TOKEN).getUser();

			assertThat(user).isNotNull();
		}

		@Test
		void shouldGetUserFromTokenWithoutUserLastname() {
			mockClaims(FIRSTNAME_CLAIM, null);

			UserProfile user = service.getUserFromToken(FAKE_TOKEN).getUser();

			assertThat(user).isNotNull();
		}

		@Test
		void shouldGetOrganisationseinheitIdsFromToken() {
			mockClaims(FIRSTNAME_CLAIM, LASTNAME_CLAIM);

			UserProfile user = service.getUserFromToken(FAKE_TOKEN).getUser();

			assertThat(user.getOrganisationseinheitIds()).isEqualTo(ORGE_IDS);
		}

		private void mockClaims(String firstnameClaim, String lastnameClaim) {
			Claims claims = mock(Claims.class);

			when(claims.get(FIRSTNAME_CLAIM, String.class)).thenReturn(firstnameClaim);
			when(claims.get(LASTNAME_CLAIM, String.class)).thenReturn(lastnameClaim);
			when(claims.get(USERID_CLAIM, String.class)).thenReturn(UserProfileTestFactory.ID.toString());

			when(jwtTokenUtil.getOrganisationseinheitIdsFromToken(any())).thenReturn(ORGE_IDS);
			when(jwtTokenUtil.getAllClaimsFromToken(any())).thenReturn(Optional.of(claims));
		}

		@Test
		void shouldThrowExceptionOnInvalidToken() throws Exception {
			doThrow(JWTVerificationException.class).when(jwtTokenUtil).verifyToken(anyString());

			assertThrows(TechnicalException.class, () -> handleToken());
		}

		@Test
		void shouldNotFilterOnInvalidToken() throws Exception {
			doThrow(JWTVerificationException.class).when(jwtTokenUtil).verifyToken(anyString());

			assertThrows(TechnicalException.class, () -> handleToken());
		}

	}

	void handleToken() {
		service.handleToken(request, FAKE_TOKEN);
	}
}