/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.postfach;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.springframework.http.MediaType;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.ResultActions;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;

import de.ozgcloud.alfa.common.command.CommandController.CommandByRelationController;
import de.ozgcloud.alfa.common.command.CommandOrder;
import de.ozgcloud.alfa.common.command.CommandTestFactory;
import de.ozgcloud.alfa.common.command.CreateCommand;
import de.ozgcloud.alfa.postfach.PostfachMailController.PostfachMailCommandController;
import de.ozgcloud.alfa.vorgang.VorgangController;
import de.ozgcloud.alfa.vorgang.VorgangHeaderTestFactory;
import de.ozgcloud.alfa.vorgang.VorgangWithEingang;
import de.ozgcloud.alfa.vorgang.VorgangWithEingangTestFactory;
import lombok.SneakyThrows;

class PostfachMailCommandControllerTest {

	@InjectMocks
	private PostfachMailCommandController controller;
	@Mock
	private PostfachMailService service;
	@Mock
	private PostfachNachrichtHelper helper;

	@Mock
	private CommandByRelationController commandByRelationController;

	@Mock
	private VorgangController vorgangController;

	private MockMvc mockMvc;

	@BeforeEach
	void initMockMvc() {
		mockMvc = MockMvcBuilders.standaloneSetup(controller).build();
	}

	@DisplayName("Create command")
	@Nested
	class TestCreateCommand {

		@BeforeEach
		void init() {
			when(service.findById(any())).thenReturn(PostfachMailTestFactory.create());
			when(commandByRelationController.createCommand(any()))
					.thenAnswer((i) -> CommandTestFactory.createBuilder().order(((CreateCommand) i.getArgument(0)).getOrder()).build());
		}

		@Test
		void shouldFindPostfachMail() {
			doRequest(CommandOrder.SEND_POSTFACH_NACHRICHT);

			verify(service).findById(PostfachMailTestFactory.NACHRICHT_ID);
		}

		@Test
		void shouldReturnResponse() throws Exception {
			doRequest(CommandOrder.SEND_POSTFACH_NACHRICHT).andExpect(status().isCreated());
		}

		@DisplayName("on RESEND_POSTFACH_MAIL order")
		@Nested
		class TestOnResendPostfachMailOrder {

			@Test
			void shouldCallServiceResend() {
				doRequest(CommandOrder.RESEND_POSTFACH_MAIL);

				verify(service).resendPostfachMail(CommandTestFactory.ID, PostfachMailTestFactory.ID);
			}

			@Test
			void shouldCallCommandByRelationController() {
				doRequest(CommandOrder.RESEND_POSTFACH_MAIL);

				verify(commandByRelationController).createCommand(any());
			}
		}

		@DisplayName("on SEND_POSTFACH_NACHRICHT order")
		@Nested
		class TestOnSendPostfachNachrichtOrder {

			@Captor
			private ArgumentCaptor<CreateCommand> createCommandCaptor;

			@BeforeEach
			void mockHelper() {
				when(helper.addPostfachAddressToBody(any(), any()))
						.thenReturn(PostfachMailTestFactory.create());
				when(vorgangController.getVorgang(anyString())).thenReturn(VorgangWithEingangTestFactory.create());
			}

			@Test
			void shouldCallVorgangController() {
				doRequest(CommandOrder.SEND_POSTFACH_NACHRICHT);

				verify(vorgangController).getVorgang(VorgangHeaderTestFactory.ID);
			}

			@Test
			void shouldCallHelperToAddPostfachAddress() {
				doRequest(CommandOrder.SEND_POSTFACH_NACHRICHT);

				verify(helper).addPostfachAddressToBody(any(PostfachMail.class), any(VorgangWithEingang.class));
			}

			@Test
			void shouldCallCommandByRelationController() {
				doRequest(CommandOrder.SEND_POSTFACH_NACHRICHT);

				verify(commandByRelationController).createCommand(createCommandCaptor.capture());
				assertThat(((PostfachMail) createCommandCaptor.getValue().getBody()).getPostfachAddress())
						.isEqualTo(PostfachAddressTestFactory.createAsMap());
			}
		}

		@SneakyThrows
		private ResultActions doRequest(CommandOrder order) {
			return mockMvc.perform(post("/api/postfachMails/" + PostfachMailTestFactory.ID + "/commands")
					.contentType(MediaType.APPLICATION_JSON)
					.content(CommandTestFactory.buildCreateCommandWithBodyContent(order, "{}")));
		}
	}
}