/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.statistic;

import static de.ozgcloud.alfa.statistic.StatisticRemoteService.*;
import static de.ozgcloud.alfa.statistic.StatisticService.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.Map;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.alfa.common.user.CurrentUserService;
import de.ozgcloud.alfa.common.user.UserRole;
import de.ozgcloud.alfa.vorgang.Vorgang.VorgangStatus;

class StatisticServiceTest {

	@Spy
	@InjectMocks
	private StatisticService service;
	@Mock
	private StatisticRemoteService remoteService;
	@Mock
	private CurrentUserService currentUserService;

	@Nested
	class TestBuildGetVorgaengeStatisticResult {

		@BeforeEach
		void beforeEach() {
			when(remoteService.buildCountByStatusResultName(any())).thenCallRealMethod();
		}

		@Test
		void shouldContainCountWiedervorlage() {
			var statistic = service.buildGetVorgaengeStatisticResult(
					Map.of(COUNT_WIEDERVORLAGE_NEXT_FRIST_RESULT_NAME, StatisticResultTestFactory.create()));

			assertThat(statistic.getWiedervorlagen()).isEqualTo(StatisticResultTestFactory.INT_VALUE);
		}

		@Test
		void shouldContainDefaultCountWiedervorlage() {
			var statistic = service.buildGetVorgaengeStatisticResult(Map.of());

			assertThat(statistic.getWiedervorlagen()).isZero();
		}

		@Test
		void shouldContainExistsWiedervorlageOverdue() {
			var statistic = service.buildGetVorgaengeStatisticResult(Map.of(
					EXISTS_WIEDERVORLAGE_OVERDUE_RESULT_NAME,
					StatisticResultTestFactory.createBuilder().boolValue(true).build()));

			assertThat(statistic.isExistsWiedervorlageOverdue()).isTrue();
		}

		@Test
		void shouldContainDefaultExistsWiedervorlageOverdue() {
			var statistic = service.buildGetVorgaengeStatisticResult(Map.of());

			assertThat(statistic.isExistsWiedervorlageOverdue()).isFalse();
		}

		@Test
		void shouldContainCountByStatusNeu() {
			var statistic = service.buildGetVorgaengeStatisticResult(Map.of(
					String.format(COUNT_VORGANG_STATUS_RESULT_NAME_FORMAT, VorgangStatus.NEU.name()),
					StatisticResultTestFactory.create()));

			assertThat(statistic.getByStatus().getNeu()).isEqualTo(StatisticResultTestFactory.INT_VALUE);
		}

		@Test
		void shouldContainCountByStatusAngenommen() {
			var statistic = service.buildGetVorgaengeStatisticResult(Map.of(
					String.format(COUNT_VORGANG_STATUS_RESULT_NAME_FORMAT, VorgangStatus.ANGENOMMEN.name()),
					StatisticResultTestFactory.create()));

			assertThat(statistic.getByStatus().getAngenommen()).isEqualTo(StatisticResultTestFactory.INT_VALUE);
		}

		@Test
		void shouldContainCountByStatusInBearbeitung() {
			var statistic = service.buildGetVorgaengeStatisticResult(Map.of(
					String.format(COUNT_VORGANG_STATUS_RESULT_NAME_FORMAT, VorgangStatus.IN_BEARBEITUNG.name()),
					StatisticResultTestFactory.create()));

			assertThat(statistic.getByStatus().getInBearbeitung()).isEqualTo(StatisticResultTestFactory.INT_VALUE);
		}

		@Test
		void shouldContainCountByStatusBeschieden() {
			var statistic = service.buildGetVorgaengeStatisticResult(Map.of(
					String.format(COUNT_VORGANG_STATUS_RESULT_NAME_FORMAT, VorgangStatus.BESCHIEDEN.name()),
					StatisticResultTestFactory.create()));

			assertThat(statistic.getByStatus().getBeschieden()).isEqualTo(StatisticResultTestFactory.INT_VALUE);
		}

		@Test
		void shouldContainCountByStatusAbgeschlossen() {
			var statistic = service.buildGetVorgaengeStatisticResult(Map.of(
					String.format(COUNT_VORGANG_STATUS_RESULT_NAME_FORMAT, VorgangStatus.ABGESCHLOSSEN.name()),
					StatisticResultTestFactory.create()));

			assertThat(statistic.getByStatus().getAbgeschlossen()).isEqualTo(StatisticResultTestFactory.INT_VALUE);
		}

		@Test
		void shouldContainCountByStatusVerworfen() {
			var statistic = service.buildGetVorgaengeStatisticResult(Map.of(
					String.format(COUNT_VORGANG_STATUS_RESULT_NAME_FORMAT, VorgangStatus.VERWORFEN.name()),
					StatisticResultTestFactory.create()));

			assertThat(statistic.getByStatus().getVerworfen()).isEqualTo(StatisticResultTestFactory.INT_VALUE);
		}

		@Test
		void shouldContainCountByStatusZuLoeschen() {
			var statistic = service.buildGetVorgaengeStatisticResult(Map.of(
					String.format(COUNT_VORGANG_STATUS_RESULT_NAME_FORMAT, VorgangStatus.ZU_LOESCHEN.name()),
					StatisticResultTestFactory.create()));

			assertThat(statistic.getByStatus().getZuLoeschen()).isEqualTo(StatisticResultTestFactory.INT_VALUE);
		}

		@Test
		void shouldContainDefaultCountByStatus() {
			var statistic = service.buildGetVorgaengeStatisticResult(Map.of());

			assertThat(statistic.getByStatus().getNeu()).isZero();
		}

		@Test
		void shouldContainUngeleseneNachrichten() {
			var statistic = service.buildGetVorgaengeStatisticResult(
					Map.of(COUNT_VORGAENGE_WITH_UNGELESENE_NACHRICHTEN_RESULT_NAME, StatisticResultTestFactory.create()));

			assertThat(statistic.getUngeleseneNachrichten()).isEqualTo(StatisticResultTestFactory.INT_VALUE);
		}
	}

	@Nested
	class TestGetCountByVorgangStatusList {

		@Test
		void shouldGetForVerwalungUser() {
			when(currentUserService.hasRole(UserRole.VERWALTUNG_USER)).thenReturn(true);

			var result = service.getCountByVorgangStatusList();

			assertThat(result).containsAll(COUNT_BY_VORGANG_STATUS_VERWALTUNG_USER);
		}

		@Test
		void shouldGetForPoststelleUser() {
			when(currentUserService.hasRole(UserRole.VERWALTUNG_USER)).thenReturn(false);
			when(currentUserService.hasRole(UserRole.VERWALTUNG_POSTSTELLE)).thenReturn(true);

			var result = service.getCountByVorgangStatusList();

			assertThat(result).containsAll(COUNT_BY_VORGANG_STATUS_POSTSTELLE_USER);
		}

		@Test
		void shouldGetForEinheitlicherAnsprechpartnerUser() {
			when(currentUserService.hasRole(UserRole.VERWALTUNG_USER)).thenReturn(false);
			when(currentUserService.hasRole(UserRole.VERWALTUNG_POSTSTELLE)).thenReturn(false);
			when(currentUserService.hasRole(UserRole.EINHEITLICHER_ANSPRECHPARTNER)).thenReturn(true);

			var result = service.getCountByVorgangStatusList();

			assertThat(result).containsAll(COUNT_BY_VORGANG_EINHEITLICHER_ANSPRECHPARTNER);
		}

		@Test
		void shouldGetForUnknownUser() {
			when(currentUserService.hasRole(UserRole.VERWALTUNG_USER)).thenReturn(false);
			when(currentUserService.hasRole(UserRole.VERWALTUNG_POSTSTELLE)).thenReturn(false);
			when(currentUserService.hasRole(UserRole.EINHEITLICHER_ANSPRECHPARTNER)).thenReturn(false);

			var result = service.getCountByVorgangStatusList();

			assertThat(result).containsAll(Collections.emptyList());
		}
	}

	@Nested
	class TestGetVorgaengeStatistics {

		private final Statistic statistic = StatisticTestFactory.create();
		private final Map<String, StatisticResult> getVorgangStatisticsResponse = Map.of(COUNT_WIEDERVORLAGE_NEXT_FRIST_RESULT_NAME,
				StatisticResultTestFactory.create());

		@BeforeEach
		void beforeEach() {
			doReturn(COUNT_BY_VORGANG_STATUS_VERWALTUNG_USER).when(service).getCountByVorgangStatusList();
			when(remoteService.getVorgaengeStatistics(COUNT_BY_VORGANG_STATUS_VERWALTUNG_USER)).thenReturn(getVorgangStatisticsResponse);
			doReturn(statistic).when(service).buildGetVorgaengeStatisticResult(getVorgangStatisticsResponse);
		}

		@Test
		void shouldCallGetCountByVorgangStatusList() {
			service.getVorgaengeStatistic();

			verify(service).getCountByVorgangStatusList();
		}

		@Test
		void shouldCallRemoteService() {
			service.getVorgaengeStatistic();

			verify(remoteService).getVorgaengeStatistics(COUNT_BY_VORGANG_STATUS_VERWALTUNG_USER);
		}

		@Test
		void shouldBuildResult() {
			service.getVorgaengeStatistic();

			verify(service).buildGetVorgaengeStatisticResult(getVorgangStatisticsResponse);
		}

		@Test
		void shouldReturnStatistic() {
			var result = service.getVorgaengeStatistic();

			assertThat(result).isEqualTo(statistic);
		}
	}
}
