/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.vorgang;

import static de.ozgcloud.alfa.common.UserProfileUrlProviderTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.List;
import java.util.Optional;
import java.util.stream.Stream;

import org.apache.commons.lang3.StringUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.hateoas.EntityModel;
import org.springframework.hateoas.IanaLinkRelations;
import org.springframework.hateoas.Link;

import de.ozgcloud.alfa.common.UserProfileUrlProvider;
import de.ozgcloud.alfa.common.user.CurrentUserService;
import de.ozgcloud.alfa.common.user.UserProfileTestFactory;
import de.ozgcloud.alfa.common.user.UserRole;
import de.ozgcloud.alfa.statistic.StatisticTestFactory;
import de.ozgcloud.alfa.vorgang.Vorgang.VorgangStatus;

class VorgangModelAssemblerTest {

	@Spy
	@InjectMocks
	private VorgangModelAssembler modelAssembler;
	@Mock
	private CurrentUserService userService;

	private UserProfileUrlProvider urlProvider = new UserProfileUrlProvider();

	private static final int PAGE_SIZE = 100;

	@DisplayName("To collectionModel")
	@Nested
	class TestCollectionModel {

		@BeforeEach
		void prepareBuilder() {
			initUserProfileUrlProvider(urlProvider);
		}

		@DisplayName("init model")
		@Nested
		class TestInitModel {

			@Test
			void shouldBeCalled() {
				var requestCriteria = FindVorgaengeRequestCriteriaTestFactory.createBuilder().searchBy(Optional.empty()).build();
				var response = VorgaengeHeaderResponse.builder().vorgaengeHeader(Collections.emptyList()).total(PAGE_SIZE).build();
				var entities = Stream.of(VorgangHeaderTestFactory.create());
				var statistic = StatisticTestFactory.create();

				modelAssembler.toCollectionModel(entities, response, requestCriteria, statistic);

				verify(modelAssembler).initModel(entities, statistic);
			}
		}

		@DisplayName("next link")
		@Nested
		class TestNextLink {

			@Test
			void shouldAddIfPagesLeft() {
				var requestCriteria = FindVorgaengeRequestCriteriaTestFactory.createBuilder().searchBy(Optional.empty()).build();
				var response = VorgaengeHeaderResponse.builder().vorgaengeHeader(Collections.emptyList()).total(PAGE_SIZE).build();

				var model = modelAssembler.toCollectionModel(Stream.of(VorgangHeaderTestFactory.create()), response, requestCriteria,
						StatisticTestFactory.create());

				assertThat(model.getLink(VorgangModelAssembler.REL_NEXT).get()).extracting(Link::getHref)
						.isEqualTo(VorgangController.PATH + "?page=1&limit=100&assignedTo=" + UserProfileTestFactory.ID.toString() + "&status=NEU");
			}

			@Test
			void shouldNotAddIfLastPage() {
				var requestCriteria = FindVorgaengeRequestCriteriaTestFactory.create();

				var response = VorgaengeHeaderResponse.builder()
						.vorgaengeHeader(List.of(VorgangHeaderTestFactory.create()))
						.total(1).build();

				var model = modelAssembler.toCollectionModel(Stream.of(VorgangHeaderTestFactory.create()), response, requestCriteria,
						StatisticTestFactory.create());

				assertThat(model.getLink(VorgangModelAssembler.REL_NEXT)).isNotPresent();
			}
		}

		@DisplayName("previous link")
		@Nested
		class TestPreviousLink {

			@Test
			void shouldAddIfNotFirstPage() {
				var requestCriteria = FindVorgaengeRequestCriteriaTestFactory.createBuilder()
						.page(1)
						.limit(1)
						.searchBy(Optional.empty())
						.build();
				var response = VorgaengeHeaderResponse.builder()
						.vorgaengeHeader(List.of(VorgangHeaderTestFactory.create()))
						.total(200)
						.build();

				var model = modelAssembler.toCollectionModel(Stream.of(VorgangHeaderTestFactory.create(), VorgangHeaderTestFactory.create()),
						response,
						requestCriteria, StatisticTestFactory.create());

				assertThat(model.getLink(VorgangModelAssembler.REL_PREVIOUS).get()).extracting(Link::getHref)
						.isEqualTo(VorgangController.PATH + "?page=0&limit=1&assignedTo=" + UserProfileTestFactory.ID.toString() + "&status=NEU");
			}

			@Test
			void shouldNotAddIfFirstPage() {
				var requestCriteria = FindVorgaengeRequestCriteriaTestFactory.createBuilder().page(0).build();
				var response = VorgaengeHeaderResponse.builder().vorgaengeHeader(Collections.emptyList()).total(PAGE_SIZE).build();

				var model = modelAssembler.toCollectionModel(Stream.of(VorgangHeaderTestFactory.create()), response, requestCriteria,
						StatisticTestFactory.create());

				assertThat(model.getLink(VorgangModelAssembler.REL_PREVIOUS)).isNotPresent();
			}
		}

		@DisplayName("add self link")
		@Nested
		class TestSelfLink {

			final FindVorgaengeHeaderRequestCriteria requestCriteria = FindVorgaengeRequestCriteriaTestFactory.createBuilder()
					.page(1)
					.limit(1)
					.build();
			final VorgaengeHeaderResponse response = VorgaengeHeaderResponse.builder()
					.vorgaengeHeader(List.of(VorgangHeaderTestFactory.create()))
					.total(200)
					.build();

			@Test
			void shouldAddLink() {
				var model = modelAssembler.toCollectionModel(Stream.of(VorgangHeaderTestFactory.create(), VorgangHeaderTestFactory.create()),
						response,
						requestCriteria, StatisticTestFactory.create());

				assertThat(model.getLink(IanaLinkRelations.SELF_VALUE)).isPresent();
				assertThat(model.getLink(IanaLinkRelations.SELF_VALUE).get()).extracting(Link::getHref).isEqualTo(VorgangController.PATH);
			}
		}
	}

	@DisplayName("Build page link")
	@Nested
	class TestBuildPageLink {

		private final String linkRel = "dummyLinkRel";
		final FindVorgaengeHeaderRequestCriteria requestCriteria = FindVorgaengeRequestCriteriaTestFactory.createBuilder().build();

		@Test
		void shouldContainsPage() {
			var link = modelAssembler.buildPageLink(0, linkRel, requestCriteria);

			assertThat(link.getHref()).contains(VorgangController.PARAM_PAGE + "=0");
		}

		@Test
		void shouldContainsLimit() {
			var link = modelAssembler.buildPageLink(0, linkRel, requestCriteria);

			assertThat(link.getHref()).contains(VorgangController.PARAM_LIMIT + "=" + FindVorgaengeRequestCriteriaTestFactory.LIMIT);
		}

		@Test
		void shouldHaveLinkRel() {
			var link = modelAssembler.buildPageLink(0, linkRel, requestCriteria);

			assertThat(link.hasRel(linkRel)).isTrue();
		}

		@DisplayName("search by")
		@Nested
		class TestSearchBy {

			@Test
			void shouldAddAsParamIfExists() {
				var link = modelAssembler.buildPageLink(0, linkRel, requestCriteria);

				assertThat(link.getHref())
						.contains(VorgangController.PARAM_SEARCH + "="
								+ FindVorgaengeRequestCriteriaTestFactory.SEARCH_BY.replace(StringUtils.SPACE, "%20"));
			}

			@Test
			void shouldNOTaddIfNotExists() {
				var link = modelAssembler.buildPageLink(0, linkRel, requestCriteria.toBuilder().searchBy(Optional.empty()).build());

				assertThat(link.getHref()).doesNotContain(VorgangController.PARAM_SEARCH);
			}
		}

		@DisplayName("assignedTo")
		@Nested
		class TestAssignedTo {

			@Test
			void shouldAddAsParamIfShouldFilter() {
				var link = modelAssembler.buildPageLink(0, linkRel, requestCriteria);

				assertThat(link.getHref()).contains(VorgangController.PARAM_ASSIGNED_TO + "=" + UserProfileTestFactory.ID);
			}

			@Test
			void shouldNOTaddIfShouldNotFilter() {
				var link = modelAssembler.buildPageLink(0, linkRel, requestCriteria.toBuilder().filterCriteria(
						FilterCriteriaTestFactory.createBuilder().filterByAssignedTo(false).build()).build());

				assertThat(link.getHref()).doesNotContain(VorgangController.PARAM_ASSIGNED_TO);
			}

			@DisplayName("getAssignedToValue")
			@Nested
			class TestGetAssignedToValue {

				@Test
				void shouldReturnValueIfExists() {
					var assignedToValue = modelAssembler.getAssignedToValue(FilterCriteriaTestFactory.create());

					assertThat(assignedToValue).isEqualTo(UserProfileTestFactory.ID.toString());
				}

				@Test
				void shouldReturnEmptyStringIfNotExists() {
					var assignedToValue = modelAssembler
							.getAssignedToValue(FilterCriteriaTestFactory.createBuilder().assignedTo(Optional.empty()).build());

					assertThat(assignedToValue).isEqualTo(StringUtils.EMPTY);
				}
			}
		}

		@DisplayName("status")
		@Nested
		class TestStatus {

			@Test
			void shouldAddAsParamIfExists() {
				var link = modelAssembler.buildPageLink(0, linkRel, requestCriteria);

				assertThat(link.getHref()).contains(VorgangController.PARAM_STATUS + "=" + FilterCriteriaTestFactory.VORGANG_STATUS);
			}

			@Test
			void shouldNOTaddIfNotExists() {
				var link = modelAssembler.buildPageLink(0, linkRel, requestCriteria.toBuilder().filterCriteria(
						FilterCriteriaTestFactory.createBuilder().status(Optional.empty()).build()).build());

				assertThat(link.getHref()).doesNotContain(VorgangController.PARAM_STATUS);
			}
		}

		@DisplayName("next wiedervorlage frist")
		@Nested
		class TestWiedervorlagen {

			@Test
			void shouldAddAsParamExists() {
				var link = modelAssembler.buildPageLink(0, linkRel, requestCriteria.toBuilder()
						.filterCriteria(FilterCriteriaTestFactory.createBuilder().hasNextWiedervorlageFrist(true).build()).build());

				assertThat(link.getHref())
						.contains(VorgangController.PARAM_NEXT_WIEDERVORLAGE_FRIST + "=" + VorgangController.PARAM_NEXT_WIEDERVORLAGE_FRIST_EXISTS);
			}

			@Test
			void shouldNOTaddIfNotExists() {
				var link = modelAssembler.buildPageLink(0, linkRel, requestCriteria.toBuilder().filterCriteria(
						FilterCriteriaTestFactory.createBuilder().hasNextWiedervorlageFrist(false).build()).build());

				assertThat(link.getHref()).doesNotContain(VorgangController.PARAM_NEXT_WIEDERVORLAGE_FRIST);
			}
		}
	}

	@Nested
	class TestLinksOnModel {

		private final UserProfileUrlProvider urlProvider = new UserProfileUrlProvider();

		@BeforeEach
		void beforeEach() {
			initUserProfileUrlProvider(urlProvider);
		}

		@Test
		void shouldHaveSelfLink() {
			var link = toModel().getLink(IanaLinkRelations.SELF_VALUE);

			assertThat(link).isPresent().get().extracting(Link::getHref).isEqualTo(VorgangController.PATH + "/" + VorgangHeaderTestFactory.ID);
		}

		@Test
		void shouldHaveVorgangMitEingangLink() {
			var link = toModel().getLink(VorgangModelAssembler.REL_VORGANG_MIT_EINGANG);

			assertThat(link).isPresent().get().extracting(Link::getHref).isEqualTo(VorgangController.PATH + "/" + VorgangHeaderTestFactory.ID);
		}

		@DisplayName("Without any roles")
		@Nested
		class WithoutRoles {

			@Test
			void shouldNotHaveWiedervorlagenLink() {
				var link = toModel().getLink(VorgangModelAssembler.REL_WIEDERVORLAGEN);

				assertThat(link).isEmpty();
			}

			@DisplayName("Should not have Vorgang-assign-links")
			@Test
			void shouldNotHaveAssignLink() {
				var link = toModel().getLink(VorgangModelAssembler.REL_VORGANG_ASSIGN);

				assertThat(link).isEmpty();
			}
		}

		@Nested
		class ForRoleVerwaltungPoststelle {

			@Test
			void shouldHaveAssignLink() {
				when(userService.hasRole(UserRole.VERWALTUNG_POSTSTELLE)).thenReturn(true);
				when(userService.hasRole(UserRole.VERWALTUNG_USER)).thenReturn(false);

				var link = toModel().getLink(VorgangModelAssembler.REL_VORGANG_ASSIGN);

				assertThat(link).isPresent().get().extracting(Link::getHref).isEqualTo(buildCommandLink());
			}

			@Test
			void shouldNotHaveAssignLink() {
				var vorgang = VorgangWithEingangTestFactory.createBuilder().status(VorgangStatus.ZU_LOESCHEN).build();
				doReturn(false).when(modelAssembler).isAssigneable(vorgang);

				var model = modelAssembler.toModel(vorgang);
				var link = model.getLink(VorgangModelAssembler.REL_VORGANG_ASSIGN);

				assertThat(link).isEmpty();
			}

		}

		@Nested
		class ForRoleVerwaltungUser {

			@BeforeEach
			void mockUserService() {
				when(userService.hasRole(UserRole.VERWALTUNG_USER)).thenReturn(true);
			}

			@Test
			void shouldHaveWiedervorlagenLink() {
				var link = toModel().getLink(VorgangModelAssembler.REL_WIEDERVORLAGEN);

				assertThat(link).isPresent().get().extracting(Link::getHref)
						.isEqualTo("/api/wiedervorlages?vorgangId=" + VorgangHeaderTestFactory.ID);
			}

			@Test
			void shouldHaveAssignLink() {
				var link = toModel().getLink(VorgangModelAssembler.REL_VORGANG_ASSIGN);

				assertThat(link).isPresent().get().extracting(Link::getHref).isEqualTo(buildCommandLink());
			}

			@Test
			void shouldNotHaveAssignLink() {
				var model = modelAssembler.toModel(VorgangWithEingangTestFactory.createBuilder().status(VorgangStatus.ZU_LOESCHEN).build());
				var link = model.getLink(VorgangModelAssembler.REL_VORGANG_ASSIGN);

				assertThat(link).isEmpty();
			}
		}

		private String buildCommandLink() {
			return String.format("/api/vorgangs/%s/relations/%s/%s/commands", VorgangHeaderTestFactory.ID,
					VorgangHeaderTestFactory.ID, VorgangHeaderTestFactory.VERSION);
		}

		private EntityModel<Vorgang> toModel() {
			return modelAssembler.toModel(VorgangWithEingangTestFactory.create());
		}
	}

	@Nested
	class TestIsAssigneable {

		@Test
		void shouldReturnTrueForVerwaltungUser() {
			when(userService.hasRole(UserRole.VERWALTUNG_USER)).thenReturn(true);
			var vorgang = VorgangHeaderTestFactory.createBuilder().status(VorgangStatus.NEU);

			assertThat(modelAssembler.isAssigneable(vorgang.build())).isTrue();
		}

		@Test
		void shouldReturnFalseForVerwaltungUser() {
			when(userService.hasRole(UserRole.VERWALTUNG_USER)).thenReturn(false);
			var vorgang = VorgangHeaderTestFactory.createBuilder().status(VorgangStatus.NEU);

			assertThat(modelAssembler.isAssigneable(vorgang.build())).isFalse();
		}

		@Test
		void shouldReturnFalseForVerwatlugnUserAndVorgangZuLoeschen() {
			when(userService.hasRole(UserRole.VERWALTUNG_USER)).thenReturn(true);
			var vorgang = VorgangHeaderTestFactory.createBuilder().status(VorgangStatus.ZU_LOESCHEN);

			assertThat(modelAssembler.isAssigneable(vorgang.build())).isFalse();
		}

		@Test
		void shouldReturnTrueForPoststelleUser() {
			when(userService.hasRole(UserRole.VERWALTUNG_USER)).thenReturn(false);
			when(userService.hasRole(UserRole.VERWALTUNG_POSTSTELLE)).thenReturn(true);
			var vorgang = VorgangHeaderTestFactory.createBuilder().status(VorgangStatus.NEU);

			assertThat(modelAssembler.isAssigneable(vorgang.build())).isTrue();
		}

		@Test
		void shouldReturnFalseForPoststelleUser() {
			when(userService.hasRole(UserRole.VERWALTUNG_USER)).thenReturn(false);
			when(userService.hasRole(UserRole.VERWALTUNG_POSTSTELLE)).thenReturn(false);
			var vorgang = VorgangHeaderTestFactory.createBuilder().status(VorgangStatus.NEU);

			assertThat(modelAssembler.isAssigneable(vorgang.build())).isFalse();
		}

		@Test
		void shouldReturnFalseForPoststelleUserAndVorgangZuLoeschen() {
			when(userService.hasRole(UserRole.VERWALTUNG_USER)).thenReturn(false);
			when(userService.hasRole(UserRole.VERWALTUNG_POSTSTELLE)).thenReturn(true);
			var vorgang = VorgangHeaderTestFactory.createBuilder().status(VorgangStatus.ZU_LOESCHEN);

			assertThat(modelAssembler.isAssigneable(vorgang.build())).isFalse();
		}
	}
}